# File defining all the constants being used.


SCALA_LANGUAGE = "scala"
PYTHON_LANGUAGE = "python"

DATABRICKS_HOST = "DATABRICKS_HOST"
DATABRICKS_TOKEN = "DATABRICKS_TOKEN"

PROJECT_ID_PLACEHOLDER_REGEX = "__PROJECT_ID_PLACEHOLDER__"
PROPHECY_URL_PLACEHOLDER_REGEX = "__PROPHECY_URL_PLACEHOLDER__"
PROJECT_RELEASE_VERSION_PLACEHOLDER_REGEX = "__PROJECT_RELEASE_VERSION_PLACEHOLDER__"
PROJECT_RELEASE_TAG_PLACEHOLDER_REGEX = "__PROJECT_FULL_RELEASE_TAG_PLACEHOLDER__"
PROJECT_URL_PLACEHOLDER_REGEX = "__PROPHECY_URL_PLACEHOLDER__"

PIPELINE_ID = "PipelineId"

PBT_FILE_NAME = "pbt_project.yml"
LANGUAGE = "language"
JOBS = "jobs"
PIPELINES = "pipelines"
GEMS = "gems"
PIPELINE_CONFIGURATIONS = "pipelineConfigurations"
BASE_PIPELINE = "basePipeline"
CONFIGURATIONS = "configurations"

JSON_EXTENSION = ".json"

DBT_COMPONENT = "DBTComponent"
SCRIPT_COMPONENT = "ScriptComponent"
PIPELINE_COMPONENT = "PipelineComponent"
COMPONENTS_LITERAL = "components"

FABRIC_ID = "fabric_id"
FABRIC_UID = "fabricUID"
SECRET_SCOPE = "secret_scope"

DBFS_FILE_STORE = "dbfs:/FileStore"
S3_FILE_STORE = "s3://"
PROPHECY_ARTIFACTS = "prophecy/artifacts"

NEW_JOB_STATE_FILE = "new_state.yml"
